@@ -80,17 +80,19 @@ mapi_item_add_recipient (const char *recipients, OlMailRecipientType type, GSLis
 static mapi_id_t
 mapi_message_item_send (MapiItem *item, GSList *attachments, GSList *recipients);
 
+#if 0
 static void
 mapi_item_debug_dump (MapiItem *item)
 {
 	printf("-----------------\n\n");
-        printf("%s(%d):%s: \n", __FILE__, __LINE__, __PRETTY_FUNCTION__);
+        printf("%s:%s: \n", G_STRLOC, G_STRFUNC);
 	printf("item->header.from : %s\n",item->header.from);
 	//Use Recipient List
 	printf("item->header.subject : %s\n",item->header.subject);
 	//printf("item->msg.body_stream : %s\n",item->msg.body_stream);
 	printf("-----------------\n\n");
 }
+#endif
 
 static void
 mapi_item_set_from(MapiItem *item, const char *from)
@@ -127,7 +129,7 @@ mapi_item_set_body_stream (MapiItem *item, CamelStream *body, MapiItemPartType p
 		if (read_size == -1) 
 			return;
 
-		stream->value = g_byte_array_append (stream->value, (char *) buf, read_size);
+		stream->value = g_byte_array_append (stream->value, buf, read_size);
 	}
 
 	switch (part_type) {
@@ -305,6 +307,11 @@ mapi_send_to (CamelTransport *transport, CamelMimeMessage *message,
 		mapi_item_set_subject(item, camel_mime_message_get_subject(message));
 	}
 
+	/*Add message threading properties */
+	item->header.references = g_strdup (camel_medium_get_header ((CamelMedium *) message, "References"));
+	item->header.in_reply_to = g_strdup (camel_medium_get_header ((CamelMedium *) message, "In-Reply-To"));
+	item->header.message_id = g_strdup (camel_medium_get_header ((CamelMedium *) message, "Message-Id"));
+
 	/* contents body */
 	multipart = (CamelMultipart *)camel_medium_get_content_object (CAMEL_MEDIUM (message));
 
@@ -399,7 +406,7 @@ mail_build_props (struct SPropValue **value, struct SPropTagArray *SPropTagArray
 	uint32_t *msgflag = g_new0 (uint32_t, 1);
 	int i=0;
 
-	props = g_new0 (struct SPropValue, 6);
+	props = g_new0 (struct SPropValue, 9);
 
 	set_SPropValue_proptag(&props[i++], PR_CONVERSATION_TOPIC_UNICODE, g_strdup (item->header.subject));
 	set_SPropValue_proptag(&props[i++], PR_NORMALIZED_SUBJECT_UNICODE, g_strdup (item->header.subject));
@@ -407,6 +414,16 @@ mail_build_props (struct SPropValue **value, struct SPropTagArray *SPropTagArray
 	*msgflag = MSGFLAG_UNSENT;
 	set_SPropValue_proptag(&props[i++], PR_MESSAGE_FLAGS, (void *)msgflag);
 
+	/* Message threading information */
+	if (item->header.references)
+		set_SPropValue_proptag(&props[i++], PR_INTERNET_REFERENCES, g_strdup (item->header.references));
+
+	if (item->header.in_reply_to)
+		set_SPropValue_proptag(&props[i++], PR_IN_REPLY_TO_ID, g_strdup (item->header.in_reply_to));
+
+	if (item->header.message_id)
+		set_SPropValue_proptag(&props[i++], PR_INTERNET_MESSAGE_ID, g_strdup (item->header.message_id));
+
 	for (l = item->msg.body_parts; l; l = l->next) {
 		ExchangeMAPIStream *stream = (ExchangeMAPIStream *) (l->data);
 		struct SBinary_short *bin = g_new0 (struct SBinary_short, 1);
